@@ -27,10 +27,14 @@
 #include <sys/time.h>
 #include <sys/resource.h>
 
+#include "lib/dshash.h"
+#include "utils/dsa.h"
+
 #include "access/hash.h"
 #include "catalog/pg_authid.h"
 #include "executor/instrument.h"
 #include "common/ip.h"
+#include "jit/jit.h"
 #include "funcapi.h"
 #include "access/twophase.h"
 #include "mb/pg_wchar.h"
@@ -53,6 +57,9 @@
 #include "utils/lsyscache.h"
 #include "utils/guc.h"
 #include "utils/guc_tables.h"
+#include "utils/memutils.h"
+#include "utils/palloc.h"
+
 
 #define MAX_BACKEND_PROCESES (MaxBackends + NUM_AUXILIARY_PROCS + max_prepared_xacts)
 #define  IntArrayGetTextDatum(x,y) intarray_get_datum(x,y)
@@ -60,7 +67,6 @@
 /* XXX: Should USAGE_EXEC reflect execution time and/or buffer usage? */
 #define USAGE_EXEC(duration)	(1.0)
 #define USAGE_INIT				(1.0)	/* including initial planning */
-#define ASSUMED_MEDIAN_INIT		(10.0)	/* initial assumed median usage */
 #define ASSUMED_LENGTH_INIT		1024	/* initial assumed mean query length */
 #define USAGE_DECREASE_FACTOR	(0.99)	/* decreased every entry_dealloc */
 #define STICKY_DECREASE_FACTOR	(0.50)	/* factor for sticky entries */
@@ -68,31 +74,34 @@
 
 #define JUMBLE_SIZE				1024	/* query serialization buffer size */
 
+#define HISTOGRAM_MAX_TIME		50000000
 #define MAX_RESPONSE_BUCKET 50
 #define INVALID_BUCKET_ID	-1
-#define MAX_REL_LEN			255
-#define MAX_BUCKETS			10
 #define TEXT_LEN			255
 #define ERROR_MESSAGE_LEN	100
+#define REL_TYPENAME_LEN	64
 #define REL_LST				10
-#define REL_LEN				1000
+#define REL_LEN				132 /* REL_TYPENAME_LEN * 2 (relname + schema) + 1
+								 * (for view indication) + 1 and dot and
+								 * string terminator */
 #define CMD_LST				10
 #define CMD_LEN				20
-#define APPLICATIONNAME_LEN	100
-#define COMMENTS_LEN        512
+#define APPLICATIONNAME_LEN	NAMEDATALEN
+#define COMMENTS_LEN        256
 #define PGSM_OVER_FLOW_MAX	10
 #define PLAN_TEXT_LEN		1024
 /* the assumption of query max nested level */
 #define DEFAULT_MAX_NESTED_LEVEL	10
 
-#define MAX_QUERY_BUF						(PGSM_QUERY_SHARED_BUFFER * 1024 * 1024)
-#define MAX_BUCKETS_MEM 					(PGSM_MAX * 1024 * 1024)
-#define BUCKETS_MEM_OVERFLOW() 				((hash_get_num_entries(pgss_hash) * sizeof(pgssEntry)) >= MAX_BUCKETS_MEM)
-#define MAX_BUCKET_ENTRIES 					(MAX_BUCKETS_MEM / sizeof(pgssEntry))
+#define MAX_QUERY_BUF						(pgsm_query_shared_buffer * 1024 * 1024)
+#define MAX_BUCKETS_MEM 					(pgsm_max * 1024 * 1024)
+#define BUCKETS_MEM_OVERFLOW() 				((hash_get_num_entries(pgsm_hash) * sizeof(pgsmEntry)) >= MAX_BUCKETS_MEM)
+#define MAX_BUCKET_ENTRIES 					(MAX_BUCKETS_MEM / sizeof(pgsmEntry))
 #define QUERY_BUFFER_OVERFLOW(x,y)  		((x + y + sizeof(uint64) + sizeof(uint64)) > MAX_QUERY_BUF)
 #define QUERY_MARGIN 						100
 #define MIN_QUERY_LEN						10
 #define SQLCODE_LEN                         20
+#define TOTAL_RELS_LENGTH					(REL_LST * REL_LEN)
 
 #if PG_VERSION_NUM >= 130000
 #define	MAX_SETTINGS                        15
@@ -102,22 +111,62 @@
 
 /* Update this if need a enum GUC with more options. */
 #define MAX_ENUM_OPTIONS 6
-typedef struct GucVariables
-{
-	enum config_type type;		/* PGC_BOOL, PGC_INT, PGC_REAL, PGC_STRING,
-								 * PGC_ENUM */
-	int			guc_variable;
-	char		guc_name[TEXT_LEN];
-	char		guc_desc[TEXT_LEN];
-	int			guc_default;
-	int			guc_min;
-	int			guc_max;
-	int			guc_unit;
-	int		   *guc_value;
-	bool		guc_restart;
-	int			n_options;
-	char		guc_options[MAX_ENUM_OPTIONS][32];
-}			GucVariable;
+
+/*
+ * API for disabling error capture ereport(ERROR,..) by PGSM's error capture hook
+ * pgsm_emit_log_hook()
+ *
+ * Use these macros as follows:
+ * 		 	PGSM_DISABLE_ERROR_CAPUTRE();
+ * 			{
+ * 				... code that might throw ereport(ERROR) ...
+ * 			}PGSM_END_DISABLE_ERROR_CAPTURE();
+ *
+ * These macros can be used to error recursion if the error gets
+ * thrown from within the function called from pgsm_emit_log_hook()
+ */
+extern volatile bool __pgsm_do_not_capture_error;
+#define PGSM_DISABLE_ERROR_CAPUTRE() \
+	do { \
+		__pgsm_do_not_capture_error = true
+
+#define PGSM_END_DISABLE_ERROR_CAPTURE() \
+	__pgsm_do_not_capture_error = false; \
+	} while (0)
+
+#define PGSM_ERROR_CAPTURE_ENABLED \
+	__pgsm_do_not_capture_error == false
+
+/*
+ * pg_stat_monitor uses the hash structure to store all query statistics
+ * except the query text, which gets stored out of line in the raw DSA area.
+ * Enabling USE_DYNAMIC_HASH uses the dshash for storing the query statistics
+ * that get created in the DSA area and can grow to any size.
+ *
+ * The only issue with using the dshash is that the newly created hash entries
+ * are explicitly locked by dshash, and its caller is required to release the lock.
+ * That works well as long as we do not want to swallow the errors thrown from
+ * dshash function. Since the lightweight locks acquired internally by dshash
+ * automatically get released by error.
+ * But throwing an error from pg_stat_monitor would mean erroring out the user query,
+ * which is not acceptable for any stat collector extension.
+ *
+ * Moreover, some of the pg_stat_monitor functions perform the sequence scan on the
+ * hash table, while the sequence scan support for dshash table is only available
+ * for PG 15 and onwards.
+ * So until we figure out the way to release the locks acquired internally by dshash
+ * in case of an error while ignoring the error at the same time, we will keep using
+ * the classic shared memory hash table.
+ */
+#ifdef USE_DYNAMIC_HASH
+#define	PGSM_HASH_TABLE	dshash_table
+#define	PGSM_HASH_TABLE_HANDLE	dshash_table_handle
+#define	PGSM_HASH_SEQ_STATUS	dshash_seq_status
+#else
+#define	PGSM_HASH_TABLE	HTAB
+#define	PGSM_HASH_TABLE_HANDLE	HTAB*
+#define	PGSM_HASH_SEQ_STATUS	HASH_SEQ_STATUS
+#endif
 
 
 #if PG_VERSION_NUM < 130000
@@ -129,29 +178,24 @@ typedef struct WalUsage
 } WalUsage;
 #endif
 
-typedef enum OVERFLOW_TARGET
-{
-	OVERFLOW_TARGET_NONE = 0,
-	OVERFLOW_TARGET_DISK
-}			OVERFLOW_TARGET;
 
-typedef enum pgssStoreKind
+typedef enum pgsmStoreKind
 {
-	PGSS_INVALID = -1,
+	PGSM_INVALID = -1,
 
 	/*
-	 * PGSS_PLAN and PGSS_EXEC must be respectively 0 and 1 as they're used to
+	 * PGSM_PLAN and PGSM_EXEC must be respectively 0 and 1 as they're used to
 	 * reference the underlying values in the arrays in the Counters struct,
-	 * and this order is required in pg_stat_statements_internal().
+	 * and this order is required in pg_stat_monitor_internal().
 	 */
-	PGSS_PARSE = 0,
-	PGSS_PLAN,
-	PGSS_EXEC,
-	PGSS_FINISHED,
-	PGSS_ERROR,
+	PGSM_PARSE = 0,
+	PGSM_PLAN,
+	PGSM_EXEC,
+	PGSM_STORE,
+	PGSM_ERROR,
 
-	PGSS_NUMKIND				/* Must be last value of this enum */
-} pgssStoreKind;
+	PGSM_NUMKIND				/* Must be last value of this enum */
+}			pgsmStoreKind;
 
 /* the assumption of query max nested level */
 #define DEFAULT_MAX_NESTED_LEVEL	10
@@ -178,20 +222,6 @@ typedef struct CallTime
 	double		sum_var_time;	/* sum of variances in execution time in msec */
 }			CallTime;
 
-/*
- * Entry type for queries hash table (query ID).
- *
- * We use a hash table to keep track of query IDs that have their
- * corresponding query text added to the query buffer (pgsm_query_shared_buffer).
- *
- * This allow us to avoid adding duplicated queries to the buffer, therefore
- * leaving more space for other queries and saving some CPU.
- */
-typedef struct pgssQueryEntry
-{
-	uint64		queryid;		/* query identifier, also the key. */
-	size_t		query_pos;		/* query location within query buffer */
-}			pgssQueryEntry;
 
 typedef struct PlanInfo
 {
@@ -200,21 +230,22 @@ typedef struct PlanInfo
 	size_t		plan_len;		/* strlen(plan_text) */
 }			PlanInfo;
 
-typedef struct pgssHashKey
+typedef struct pgsmHashKey
 {
 	uint64		bucket_id;		/* bucket number */
 	uint64		queryid;		/* query identifier */
-	uint64		userid;			/* user OID */
-	uint64		dbid;			/* database OID */
-	uint64		ip;				/* client ip address */
 	uint64		planid;			/* plan identifier */
 	uint64		appid;			/* hash of application name */
-	uint64		toplevel;		/* query executed at top level */
-} pgssHashKey;
+	Oid			userid;			/* user OID */
+	Oid			dbid;			/* database OID */
+	uint32		ip;				/* client ip address */
+	bool		toplevel;		/* query executed at top level */
+}			pgsmHashKey;
 
 typedef struct QueryInfo
 {
 	uint64		parentid;		/* parent queryid of current query */
+	dsa_pointer parent_query;
 	int64		type;			/* type of query, options are query, info,
 								 * warning, error, fatal */
 	char		application_name[APPLICATIONNAME_LEN];
@@ -255,12 +286,49 @@ typedef struct Blocks
 	int64		temp_blks_written;	/* # of temp blocks written */
 	double		blk_read_time;	/* time spent reading, in msec */
 	double		blk_write_time; /* time spent writing, in msec */
+
+	double		temp_blk_read_time; /* time spent reading temp blocks, in msec */
+	double		temp_blk_write_time;	/* time spent writing temp blocks, in
+										 * msec */
+
+	/*
+	 * Variables for local entry. The values to be passed to pgsm_update_entry
+	 * from pgsm_store.
+	 */
+	instr_time	instr_blk_read_time;	/* time spent reading blocks */
+	instr_time	instr_blk_write_time;	/* time spent writing blocks */
+	instr_time	instr_temp_blk_read_time;	/* time spent reading temp blocks */
+	instr_time	instr_temp_blk_write_time;	/* time spent writing temp blocks */
 }			Blocks;
 
+typedef struct JitInfo
+{
+	int64		jit_functions;	/* total number of JIT functions emitted */
+	double		jit_generation_time;	/* total time to generate jit code */
+	int64		jit_inlining_count; /* number of times inlining time has been
+									 * > 0 */
+	double		jit_inlining_time;	/* total time to inline jit code */
+	int64		jit_optimization_count; /* number of times optimization time
+										 * has been > 0 */
+	double		jit_optimization_time;	/* total time to optimize jit code */
+	int64		jit_emission_count; /* number of times emission time has been
+									 * > 0 */
+	double		jit_emission_time;	/* total time to emit jit code */
+
+	/*
+	 * Variables for local entry. The values to be passed to pgsm_update_entry
+	 * from pgsm_store.
+	 */
+	instr_time	instr_generation_counter;	/* generation counter */
+	instr_time	instr_inlining_counter; /* inlining counter */
+	instr_time	instr_optimization_counter; /* optimization counter */
+	instr_time	instr_emission_counter; /* emission counter */
+}			JitInfo;
+
 typedef struct SysInfo
 {
-	float		utime;			/* user cpu time */
-	float		stime;			/* system cpu time */
+	double		utime;			/* user cpu time */
+	double		stime;			/* system cpu time */
 }			SysInfo;
 
 typedef struct Wal_Usage
@@ -272,7 +340,6 @@ typedef struct Wal_Usage
 
 typedef struct Counters
 {
-	uint64		bucket_id;		/* bucket id */
 	Calls		calls;
 	QueryInfo	info;
 	CallTime	time;
@@ -283,11 +350,11 @@ typedef struct Counters
 
 	Blocks		blocks;
 	SysInfo		sysinfo;
+	JitInfo		jitinfo;
 	ErrorInfo	error;
 	Wal_Usage	walusage;
 	int			resp_calls[MAX_RESPONSE_BUCKET];	/* execution time's in
 													 * msec */
-	uint64		state;			/* query state */
 } Counters;
 
 /* Some global structure to get the cpu usage, really don't like the idea of global variable */
@@ -295,59 +362,55 @@ typedef struct Counters
 /*
  * Statistics per statement
  */
-typedef struct pgssEntry
+typedef struct pgsmEntry
 {
-	pgssHashKey key;			/* hash key of entry - MUST BE FIRST */
+	pgsmHashKey key;			/* hash key of entry - MUST BE FIRST */
+	uint64		pgsm_query_id;	/* pgsm generate normalized query hash */
+	char		datname[NAMEDATALEN];	/* database name */
+	char		username[NAMEDATALEN];	/* user name */
 	Counters	counters;		/* the statistics for this query */
 	int			encoding;		/* query text encoding */
 	slock_t		mutex;			/* protects the counters only */
-	size_t		query_pos;		/* query location within query buffer */
-} pgssEntry;
+	union
+	{
+		dsa_pointer query_pos;	/* query location within query buffer */
+		char	   *query_pointer;
+	}			query_text;
+}			pgsmEntry;
 
 /*
  * Global shared state
  */
-typedef struct pgssSharedState
+typedef struct pgsmSharedState
 {
 	LWLock	   *lock;			/* protects hashtable search/modification */
-	double		cur_median_usage;	/* current median usage in hashtable */
 	slock_t		mutex;			/* protects following fields only: */
-	Size		extent;			/* current extent of query file */
-	int64		n_writers;		/* number of active writers to query file */
 	pg_atomic_uint64 current_wbucket;
 	pg_atomic_uint64 prev_bucket_sec;
-	uint64		bucket_entry[MAX_BUCKETS];
-	char		bucket_start_time[MAX_BUCKETS][60]; /* start time of the
-													 * bucket */
-	LWLock	   *errors_lock;	/* protects errors hashtable
-								 * search/modification */
+	int			hash_tranche_id;
+	void	   *raw_dsa_area;	/* DSA area pointer to store query texts.
+								 * dshash also lives in this memory when
+								 * USE_DYNAMIC_HASH is enabled */
+	PGSM_HASH_TABLE_HANDLE hash_handle;
 
 	/*
-	 * These variables are used when pgsm_overflow_target is ON.
-	 *
-	 * overflow is set to true when the query buffer overflows.
-	 *
-	 * n_bucket_cycles counts the number of times we changed bucket since the
-	 * query buffer overflowed. When it reaches pgsm_max_buckets we remove the
-	 * dump file, also reset the counter.
-	 *
-	 * This allows us to avoid having a large file on disk that would also
-	 * slowdown queries to the pg_stat_monitor view.
+	 * hash table handle. can be either classic shared memory hash or dshash
+	 * (if we are using USE_DYNAMIC_HASH)
 	 */
-	bool		overflow;
-	size_t		n_bucket_cycles;
-} pgssSharedState;
 
-#define ResetSharedState(x) \
-do { \
-		x->cur_median_usage = ASSUMED_MEDIAN_INIT; \
-		x->cur_median_usage = ASSUMED_MEDIAN_INIT; \
-		x->n_writers = 0; \
-		pg_atomic_init_u64(&x->current_wbucket, 0); \
-		pg_atomic_init_u64(&x->prev_bucket_sec, 0); \
-		memset(&x->bucket_entry, 0, MAX_BUCKETS * sizeof(uint64)); \
-} while(0)
+	bool		pgsm_oom;
+	TimestampTz bucket_start_time[]; /* start time of the bucket */
+}			pgsmSharedState;
 
+typedef struct pgsmLocalState
+{
+	pgsmSharedState *shared_pgsmState;
+	dsa_area   *dsa;			/* local dsa area for backend attached to the
+								 * dsa area created by postmaster at startup. */
+	PGSM_HASH_TABLE *shared_hash;
+    MemoryContext pgsm_mem_cxt;
+
+}			pgsmLocalState;
 
 #if PG_VERSION_NUM < 140000
 /*
@@ -385,45 +448,35 @@ typedef struct JumbleState
 } JumbleState;
 #endif
 
-/* Links to shared memory state */
-
-bool		SaveQueryText(uint64 bucketid,
-						  uint64 queryid,
-						  unsigned char *buf,
-						  const char *query,
-						  uint64 query_len,
-						  size_t *query_pos);
-
 /* guc.c */
 void		init_guc(void);
-GucVariable *get_conf(int i);
 
 /* hash_create.c */
+dsa_area   *get_dsa_area_for_query_text(void);
+PGSM_HASH_TABLE *get_pgsmHash(void);
+
+void		pgsm_attach_shmem(void);
 bool		IsHashInitialize(void);
-void		pgss_shmem_startup(void);
-void		pgss_shmem_shutdown(int code, Datum arg);
+bool		IsSystemOOM(void);
+void		pgsm_shmem_startup(void);
+void		pgsm_shmem_shutdown(int code, Datum arg);
 int			pgsm_get_bucket_size(void);
-pgssSharedState *pgsm_get_ss(void);
-HTAB	   *pgsm_get_plan_hash(void);
-HTAB	   *pgsm_get_hash(void);
-HTAB	   *pgsm_get_query_hash(void);
-HTAB	   *pgsm_get_plan_hash(void);
-void		hash_entry_reset(void);
-void		hash_query_entryies_reset(void);
+pgsmSharedState *pgsm_get_ss(void);
 void		hash_query_entries();
 void		hash_query_entry_dealloc(int new_bucket_id, int old_bucket_id, unsigned char *query_buffer[]);
 void		hash_entry_dealloc(int new_bucket_id, int old_bucket_id, unsigned char *query_buffer);
-pgssEntry  *hash_entry_alloc(pgssSharedState *pgss, pgssHashKey *key, int encoding);
-Size		hash_memsize(void);
-
-int			read_query_buffer(int bucket_id, uint64 queryid, char *query_txt, size_t pos);
-uint64		read_query(unsigned char *buf, uint64 queryid, char *query, size_t pos);
-void		pgss_startup(void);
-void		set_qbuf(unsigned char *);
+pgsmEntry  *hash_entry_alloc(pgsmSharedState * pgsm, pgsmHashKey * key, int encoding);
+Size		pgsm_ShmemSize(void);
+void		pgsm_startup(void);
 
 /* hash_query.c */
-void		pgss_startup(void);
+void		pgsm_startup(void);
+MemoryContext GetPgsmMemoryContext(void);
 
+/* guc.c */
+void		init_guc(void);
+
+/* GUC variables*/
 /*---- GUC variables ----*/
 typedef enum
 {
@@ -439,90 +492,29 @@ static const struct config_enum_entry track_options[] =
 	{NULL, 0, false}
 };
 
-#define PGSM_MAX get_conf(0)->guc_variable
-#define PGSM_QUERY_MAX_LEN get_conf(1)->guc_variable
-#define PGSM_TRACK_UTILITY get_conf(2)->guc_variable
-#define PGSM_NORMALIZED_QUERY get_conf(3)->guc_variable
-#define PGSM_MAX_BUCKETS get_conf(4)->guc_variable
-#define PGSM_BUCKET_TIME get_conf(5)->guc_variable
-#define PGSM_HISTOGRAM_MIN get_conf(6)->guc_variable
-#define PGSM_HISTOGRAM_MAX get_conf(7)->guc_variable
-#define PGSM_HISTOGRAM_BUCKETS get_conf(8)->guc_variable
-#define PGSM_QUERY_SHARED_BUFFER get_conf(9)->guc_variable
-#define PGSM_OVERFLOW_TARGET get_conf(10)->guc_variable
-#define PGSM_QUERY_PLAN get_conf(11)->guc_variable
-#define PGSM_TRACK get_conf(12)->guc_variable
-#define PGSM_EXTRACT_COMMENTS get_conf(13)->guc_variable
-#define PGSM_TRACK_PLANNING get_conf(14)->guc_variable
-
-
-/*---- Benchmarking ----*/
-#ifdef BENCHMARK
-/*
- * These enumerator values are used as index in the hook stats array.
- * STATS_START and STATS_END are used only to delimit the range.
- * STATS_END is also the length of the valid items in the enum.
- */
-enum pg_hook_stats_id
+typedef enum
 {
-	STATS_START = -1,
-	STATS_PGSS_POST_PARSE_ANALYZE,
-	STATS_PGSS_EXECUTORSTART,
-	STATS_PGSS_EXECUTORUN,
-	STATS_PGSS_EXECUTORFINISH,
-	STATS_PGSS_EXECUTOREND,
-	STATS_PGSS_PROCESSUTILITY,
-#if PG_VERSION_NUM >= 130000
-	STATS_PGSS_PLANNER_HOOK,
-#endif
-	STATS_PGSM_EMIT_LOG_HOOK,
-	STATS_PGSS_EXECUTORCHECKPERMS,
-	STATS_END
-};
+	HISTOGRAM_START,
+	HISTOGRAM_END,
+	HISTOGRAM_COUNT
+}			HistogramTimingType;
 
-/* Hold time to execute statistics for a hook. */
-struct pg_hook_stats_t
-{
-	char		hook_name[64];
-	double		min_time;
-	double		max_time;
-	double		total_time;
-	uint64		ncalls;
-};
-
-#define HOOK_STATS_SIZE MAXALIGN((size_t)STATS_END * sizeof(struct pg_hook_stats_t))
-
-/* Allocate a pg_hook_stats_t array of size HOOK_STATS_SIZE on shared memory. */
-void		init_hook_stats(void);
-
-/* Update hook time execution statistics. */
-void		update_hook_stats(enum pg_hook_stats_id hook_id, double time_elapsed);
-
-/*
- * Macro used to declare a hook function:
- * Example:
- *    DECLARE_HOOK(void my_hook, const char *query, size_t length);
- * Will expand to:
- *    static void my_hook(const char *query, size_t length);
- *    static void my_hook_benchmark(const char *query, size_t length);
- */
-#define DECLARE_HOOK(hook, ...) \
-        static hook(__VA_ARGS__); \
-        static hook##_benchmark(__VA_ARGS__);
-
-/*
- * Macro used to wrap a hook when pg_stat_monitor is compiled with -DBENCHMARK.
- *
- * It is intended to be used as follows in _PG_init():
- *     pg_hook_function = HOOK(my_hook_function);
- * Then, if pg_stat_monitor is compiled with -DBENCHMARK this will expand to:
- *     pg_hook_name = my_hook_function_benchmark;
- * Otherwise it will simple expand to:
- *     pg_hook_name = my_hook_function;
- */
-#define HOOK(name) name##_benchmark
-
-#else							/* #ifdef BENCHMARK */
+extern int	pgsm_max;
+extern int	pgsm_query_max_len;
+extern int	pgsm_bucket_time;
+extern int	pgsm_max_buckets;
+extern int	pgsm_histogram_buckets;
+extern double pgsm_histogram_min;
+extern double pgsm_histogram_max;
+extern int	pgsm_query_shared_buffer;
+extern bool pgsm_track_planning;
+extern bool pgsm_extract_comments;
+extern bool pgsm_enable_query_plan;
+extern bool pgsm_enable_overflow;
+extern bool pgsm_normalized_query;
+extern bool pgsm_track_utility;
+extern bool pgsm_enable_pgsm_query_id;
+extern int	pgsm_track;
 
 #define DECLARE_HOOK(hook, ...) \
         static hook(__VA_ARGS__);
@@ -530,4 +522,9 @@ void		update_hook_stats(enum pg_hook_stats_id hook_id, double time_elapsed);
 #define HOOK_STATS_SIZE 0
 #endif
 
-#endif
+void	   *pgsm_hash_find_or_insert(PGSM_HASH_TABLE * shared_hash, pgsmHashKey * key, bool *found);
+void	   *pgsm_hash_find(PGSM_HASH_TABLE * shared_hash, pgsmHashKey * key, bool *found);
+void		pgsm_hash_seq_init(PGSM_HASH_SEQ_STATUS * hstat, PGSM_HASH_TABLE * shared_hash, bool lock);
+void	   *pgsm_hash_seq_next(PGSM_HASH_SEQ_STATUS * hstat);
+void		pgsm_hash_seq_term(PGSM_HASH_SEQ_STATUS * hstat);
+void		pgsm_hash_delete_current(PGSM_HASH_SEQ_STATUS * hstat, PGSM_HASH_TABLE * shared_hash, void *key);
